/* lgs.c
 * loest ein lineares Gleichungssystem und gibt zusaetzliche Infos
 * erstellt: Mit Okt 27 15:09:30 CEST 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */

#include "matrix.h"
#include "nmmatrix.h"
#include "lgs.h"
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>


/* Einbinden der externen Klassenobjekte */

extern matrix nm_matrix;
extern matrix band_matrix;
extern matrix a06__matrix;


/* Deklaration einiger Hilfsfunktionen */


/*
 *  1.) allgemeines Gaussverfahren 
 */

static int gauss_lrZerlegungUndDeterminante(linearesSystem *lgs,
                                      pivotStrategieFunktion pivotStrategie);
static int gauss_vorwaertssubstitution(linearesSystem *lgs);
static int gauss_rueckwaertssubstitution(linearesSystem *lgs);


/*
 *  2.) vereinfachtes Gaussverfahren fuer Tridiagonalmatrizen, die die
 *      Diagonalstrategie erlauben
 */

static int tgauss_lrZerlegungUndDeterminante(linearesSystem *lgs);
static int tgauss_vorwaertssubstitution(linearesSystem *lgs);
static int tgauss_rueckwaertssubstitution(linearesSystem *lgs);


/* 
 *  3.) vereinfachtes CholeskyVerfahren fuer Aufgabe 06 - Matrizen 
 */

static int chol06_lrZerlegungUndDeterminante(linearesSystem *lgs);
static int chol06_vorwaertssubstitution(linearesSystem *lgs);
static int chol06_rueckwaertssubstitution(linearesSystem *lgs);


/* ======================================================================== 
 * Implementation   
 */ 


int lrZerlegungUndDeterminanteUndLoesen(linearesSystem* lgs, 
                                        pivotStrategieFunktion pivotStrategie)
{
  int i, j;
  nmmatrixschablone schablone;
  enum moeglicheVerfahren {keins, gauss, tgauss, chol06};
  enum moeglicheVerfahren verfahren;

  /* A und b pruefen */

  if ((!lgs->a) || (!lgs->b))
    return 0;
  if (lgs->a->aktion->breite(lgs->a) != lgs->a->aktion->hoehe(lgs->a))
    return 0;
  if (lgs->b->aktion->breite(lgs->b) != 1)   
    return 0;
  if (lgs->a->aktion->breite(lgs->a) != lgs->b->aktion->hoehe(lgs->b))
    return 0;

  /* anzuwendendes Verfahren bestimmen */

  verfahren = keins;
      
  if (lgs->a->aktion->klasse(lgs->a) == &nm_matrix) 
                                                    /* Falls A nXm-Matrix */
    verfahren = gauss;              /* waehle allgemeines Gauss-Verfahren */

  if ( (lgs->a->aktion->klasse(lgs->a) == &band_matrix) 
       && (lgs->a->aktion->bandbreite(lgs->a) == 1)
       && ((pivotStrategie == diagonalStrategie) 
           || (pivotStrategie == NULL) ))

    /* Falls A Bandmatrix und tridiagonal ist und der Benutzer keine
     * Pivotstrategie ausser Diagonalstrategie angegeben hat, verwende 
     * Algorithmus von Blatt 1 Aufgabe 3. 
     */

    verfahren = tgauss;

  if (lgs->a->aktion->klasse(lgs->a) == &a06__matrix) 

    /* Falls A eine Matrix von Typ Aufgabe 06 ist, verwende dazugehoerigen
     * Algorithmus unabhaengig von der eventuell angegebenen Pivotstrate-
     * gie.
     */

    verfahren = chol06;

  /* gewaehltes Verfahren anwenden */

  switch (verfahren) {
    
    case gauss:  
                 if (pivotStrategie == NULL) {
                   fprintf(stderr, "Gauss-Verfahren erfodert die Angabe "
                                   "einer Pivotstrategie.\n");
                   exit(1);
                 }
 
                 /* Schritte des allgemeinen Gaussverfahrens: */

                 if (!gauss_lrZerlegungUndDeterminante(lgs, pivotStrategie))
                   return 0;               /* Fehler: Matrix num. singulaer */
                 gauss_vorwaertssubstitution(lgs);
                 gauss_rueckwaertssubstitution(lgs);

                 break;
    
    case tgauss: 
                 /* Schritte des angepassten Gaussverfahrens: */

                 if (!tgauss_lrZerlegungUndDeterminante(lgs))
                   return 0;                  /* Fehler: Matrix singulaer */
                 tgauss_vorwaertssubstitution(lgs);
                 tgauss_rueckwaertssubstitution(lgs);
      
                 break;

    case chol06: 
                 /* Schritte des angepassten Choleskyverfahrens: */

                 if (!chol06_lrZerlegungUndDeterminante(lgs))
                   return 0;                  /* Fehler: Matrix singulaer */
                 chol06_vorwaertssubstitution(lgs);
                 chol06_rueckwaertssubstitution(lgs);
      
                 break;

    default:
                 fprintf(stderr, "Fuer Matrixtyp %s kein Loesungsverfahren "
                                 " verfuegbar\n", 
                                 lgs->a->aktion->klasse(lgs->a)
                                   ->aktion->name(lgs->a->aktion
                                                   ->klasse(lgs->a)));
                 exit(1);
                 break;
  }
 
  /* Kompakt gespeicherte LR-Zerlegung in Matrizen l und r kopieren */

  /* zu erledigen: je nach Verfahren unterschiedlich kopieren  oder
   *               besser: weitere Matrixtypen definieren, die auf die
   *               kompakte Darstellung durchgreifen 
   */
    
  schablone.n = lgs->a->aktion->hoehe(lgs->a);
  schablone.m = lgs->a->aktion->breite(lgs->a);
  schablone.init = 0;
  schablone.name = "R";
  lgs->r = nm_matrix.aktion->kreieren(&schablone, NULL);
  
  switch (verfahren) {

  case gauss:
  case tgauss:

  schablone.name = "L";
  lgs->l = nm_matrix.aktion->kreieren(&schablone, NULL);

  for (i=1; i<=schablone.n; i++) {
    for (j=1; j<=schablone.m; j++) {
      if (i>j) {                                 /* unterhalb der Diagonale */
        lgs->l->aktion->setzen(lgs->l, i, j,     /* l(i,j) = a(i,j) */
          lgs->a->aktion->amIndex(lgs->a, i, j));
        lgs->r->aktion->setzen(lgs->r, i, j, 0); /* r(i,j) = 0; */
      }
      if (i==j) {                                /* in der Diagonale */
        lgs->l->aktion->setzen(lgs->l, i, j, 1); /* l(i,j) = 1 */
        lgs->r->aktion->setzen(lgs->r, i, j,     /* r(i,j) = a(i,j); */
          lgs->a->aktion->amIndex(lgs->a, i, j));
      }
      if (i<j) {                                 /* oberhalb der Diagonale */
        lgs->l->aktion->setzen(lgs->l, i, j, 0); /* l(i,j) = 0 */
        lgs->r->aktion->setzen(lgs->r, i, j,     /* r(i,j) = a(i,j); */
          lgs->a->aktion->amIndex(lgs->a, i, j));
      }
    }
  }
  break;

  case chol06:   /* L wurde schon explizit gemacht, da es nicht in 
                  * A gespeichert werden konnte (neue Eintraege) 
                  */

  for (i=1; i<=schablone.n; i++) {
    for (j=1; j<=schablone.m; j++) {
      lgs->r->aktion->setzen(lgs->r, i, j,     /* r(i,j) = l(j,i); */
        lgs->l->aktion->amIndex(lgs->l, j, i));
    }
  }
  break;

  default:
                 fprintf(stderr, "Fuer Matrixtyp %s konnte die kompakte "
                                 "Zerlegung nicht aufgeloest werden.\n", 
                                 lgs->a->aktion->klasse(lgs->a)
                                   ->aktion->name(lgs->a->aktion
                                                   ->klasse(lgs->a)));
                 exit(1);
                 break;
  break;
  }

  return 1; /* Erfolg */
}


/* ----------------------------------------------------------------------
 *
 *  1.) allgemeines Gaussverfahren 
 */


/* 1. Schritt: LR-Zerlegung */

static int gauss_lrZerlegungUndDeterminante(linearesSystem *lgs,
                                      pivotStrategieFunktion pivotStrategie)
{
  int i, j, k, n, pk;
  double h;                       /* Hilfsvariable zum Tauschen */
  nmmatrixschablone schablone;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Vektor p";
  schablone.init = 1;
  lgs->p = nm_matrix.aktion->kreieren(&schablone, NULL);
  lgs->det = 1.0;
  for (k=1; k<=(n-1); k++) {
    pk = pivotStrategie(lgs->a, k);  /* Pivotzeile bestimmen */
    if (!pk) {                       /* Matrix singulaer? */
      return 0;                      /* Fehler anzeigen */
    }
    lgs->p->aktion->setzen(lgs->p,k,1,pk);     /* und in Vektor p eintragen */
    if (pk != k) {                   /* Sollen Zeilen getauscht werden? */ 
      lgs->det *= -1;                /* Vorzeichen der Determinate wechseln */
      for (j=1; j<=n; j++) {         /* Tauschen */
        h = lgs->a->aktion->amIndex(lgs->a, k, j);
        lgs->a->aktion->setzen(lgs->a, k, j, 
          lgs->a->aktion->amIndex(lgs->a, pk, j));
        lgs->a->aktion->setzen(lgs->a, pk, j, h);
      }
    }
    lgs->det *=                      /* Determinante fortsetzen*/ 
      lgs->a->aktion->amIndex(lgs->a, k, k); 
    for (i=k+1; i<=n; i++) {
      lgs->a->aktion->setzen(lgs->a, i, k,     /* l(i,k) bestimmen */
          lgs->a->aktion->amIndex(lgs->a, i, k)
          / lgs->a->aktion->amIndex(lgs->a, k, k));
      for (j=k+1; j<=n; j++) {
        lgs->a->aktion->setzen(lgs->a, i, j,   /* vordersten Koeffizienten */
            lgs->a->aktion->amIndex(lgs->a, i, j) /* auf Null bringen */
            - lgs->a->aktion->amIndex(lgs->a, i, k)
              * lgs->a->aktion->amIndex(lgs->a, k, j));
      }
    }
  }
  h = lgs->a->aktion->amIndex(lgs->a, n, n);   /* Eintrag a(n,n) */
  if (fabs(h) < EPSILON) 
    return 0;                        /* Fehler anzeigen */
  lgs->det *= h;                     /* Determinatenberechnung abschliessen */
  return 1;                          /* Erfolg melden */
}



/* 2. Schritt des Gauss-Verfahrens */

static int gauss_vorwaertssubstitution(linearesSystem *lgs)
{
  int i, j, k, n, pk;
  double h;
  nmmatrixschablone schablone;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Vektor c";
  schablone.init = 1;
  lgs->c = nm_matrix.aktion->kreieren(&schablone, NULL);  /* neuer Vektor c */
  for (k=1; k<=(n-1); k++) {
    pk = lgs->p->aktion->amIndex(lgs->p, k, 1);
    if (pk != k) {
      h = lgs->b->aktion->amIndex(lgs->b, k, 1);   /* Tausch */
      lgs->b->aktion->setzen(lgs->b, k, 1, 
        lgs->b->aktion->amIndex(lgs->b, pk, 1));
      lgs->b->aktion->setzen(lgs->b, pk, 1, h);
    }
  }
  for (i=1; i<=n; i++) {
    lgs->c->aktion->setzen(lgs->c, i, 1,           /* c(i) = b(i)           */
      lgs->b->aktion->amIndex(lgs->b, i, 1));
    for (j=1; j<=(i-1); j++) {
      lgs->c->aktion->setzen(lgs->c, i, 1,         /* c(i) =                */
        lgs->c->aktion->amIndex(lgs->c, i, 1)      /*        c(i)           */
        - lgs->a->aktion->amIndex(lgs->a, i, j)    /*        - a(i,j)       */
          * lgs->c->aktion->amIndex(lgs->c, j, 1));/*          * c(j)       */
    }
  }
  return 1;   /* Erfolg */
}


/* 3. Schritt des Gauss-Verfahrens */


/* Achtung: Eintraege lgs->a(k,k) muessen ungleich null sein, da
 *          durch sie ohne Kontrolle geteilt wird. 
 */

static int gauss_rueckwaertssubstitution(linearesSystem *lgs)
{
  int i, k, n;
  double s;
  nmmatrixschablone schablone;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Loesungsvektor x";
  schablone.init = 1;
  lgs->x = nm_matrix.aktion->kreieren(&schablone, NULL);
  for (i=n; i>=1; i--) {                       /* i von Zeile n bis 1  */
    s = lgs->c->aktion->amIndex(lgs->c, i, 1);       /* c(i) +         */
    for (k=(i+1); k<=n; k++) {                       /* Summe          */
      s += lgs->a->aktion->amIndex(lgs->a, i, k)     /*   a(i,k)       */
           * lgs->x->aktion->amIndex(lgs->x, k, 1);  /*   * x(k) von   */
    }                                                /* k=i+1 bis n    */
    s /= lgs->a->aktion->amIndex(lgs->a, i, i); /* Zeile normalisieren */
    lgs->x->aktion->setzen(lgs->x, i, 1, -s);   /* und x(i) ist freigestellt */
  }
  return 1;    /* Erfolg */
}


/* ----------------------------------------------------------------------
 *
 *  2.) vereinfachtes Gaussverfahren fuer Tridiagonalmatrizen, die die
 *      Diagonalstrategie erlauben
 */


/* 1. Schritt: LR-Zerlegung */

/* Hinweis: Diese Funktion kann auch mit Bandmatrizen arbeiten, obwohl
 *          zeilen- und spaltenweise angesprochen wird. Die Bandmatrizen
 *          lenken die Zugriffe selbstaendig auf die richtigen Diagonal-
 *          vektoren um. 
 */

static int tgauss_lrZerlegungUndDeterminante(linearesSystem *lgs)
{
  int i, j, k, n;
  nmmatrixschablone schablone;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "L";
  schablone.init = 1;
  lgs->det = 1.0;
  for (k=1; k<=(n-1); k++) {
    lgs->det *=                      /* Determinante fortsetzen*/ 
      lgs->a->aktion->amIndex(lgs->a, k, k); 
    /* l(k+1,k) berechnen */
    lgs->a->aktion->setzen(lgs->a, k+1, k,
      lgs->a->aktion->amIndex(lgs->a, k+1, k)
      / lgs->a->aktion->amIndex(lgs->a, k, k));
    i = k+1;
    j = k+1;
    lgs->a->aktion->setzen(lgs->a, i, j,   /* vordersten Koeffizienten */
        lgs->a->aktion->amIndex(lgs->a, i, j) /* auf Null bringen */
        - lgs->a->aktion->amIndex(lgs->a, i, k)
          * lgs->a->aktion->amIndex(lgs->a, k, j));
  }
  lgs->det *=                        /* Determinatenberechnung abschliessen */
    lgs->a->aktion->amIndex(lgs->a, n, n); 
  return 1;                          /* Erfolg melden */
}


/* 2. Schritt des vereinfachten Gauss-Verfahrens fuer Tridiagonalmatrizen 
 * 
 * Art der Vereinfachung: Nur Elemente in der Tridiagonal muessen
 *                        beruecksichtigt werden.
 */


static int tgauss_vorwaertssubstitution(linearesSystem *lgs)
{
  int i, j, n;
  nmmatrixschablone schablone;
  double s;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Vektor c";
  schablone.init = 1;
  lgs->c = nm_matrix.aktion->kreieren(&schablone, NULL);  /* neuer Vektor c */
  for (i=1; i<=n; i++) {
    s = lgs->b->aktion->amIndex(lgs->b, i, 1);     /* s = b(i)             */
    j = i-1;
    if (j>=1)
      s -= lgs->a->aktion->amIndex(lgs->a, i, j)   /* s = s - a(i,j)       */
           * lgs->c->aktion->amIndex(lgs->c, j, 1);/*         * c(j)       */
    lgs->c->aktion->setzen(lgs->c, i, 1, s);       /* c(i) = s             */
  }
  return 1;   /* Erfolg */
}


/* 3. Schritt des vereinfachten Gauss-Verfahrens fuer Tridiagonalmatrizen 
 * 
 * Art der Vereinfachung: Nur Elemente in der Tridiagonal muessen
 *                        beruecksichtigt werden.
 */


/* Achtung: Eintraege lgs->a(k,k) muessen ungleich null sein, da
 *          durch sie ohne Kontrolle geteilt wird. 
 */

static int tgauss_rueckwaertssubstitution(linearesSystem *lgs)
{
  int i, k, n;
  double s;
  nmmatrixschablone schablone;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Loesungsvektor x";
  schablone.init = 1;
  lgs->x = nm_matrix.aktion->kreieren(&schablone, NULL);
  for (i=n; i>=1; i--) {                       /* i von Zeile n bis 1  */
    s = lgs->c->aktion->amIndex(lgs->c, i, 1);       /* c(i) +         */
    k = i+1;
    if (k<=n)
      s += lgs->a->aktion->amIndex(lgs->a, i, k)     /*   a(i,k)       */
           * lgs->x->aktion->amIndex(lgs->x, k, 1);  /*   * x(k)       */
    s /= lgs->a->aktion->amIndex(lgs->a, i, i); /* Zeile normalisieren */
    lgs->x->aktion->setzen(lgs->x, i, 1, -s);   /* und x(i) ist freigestellt */
  }
  return 1;    /* Erfolg */
}


/* ----------------------------------------------------------------------
 *
 *  3.) vereinfachtes CholeskyVerfahren fuer Aufgabe 06 - Matrizen 
 */


/* 1. Schritt: LR-Zerlegung */

static int chol06_lrZerlegungUndDeterminante(linearesSystem *lgs)
{
  int k, n;
  nmmatrixschablone schablone;
  matrix *l, *a, *b, *c;
  double lkk, lk1k, bk, ck;         /* Hilfsvariablen fuer weniger Zugriffe */
  double lnk, h, ck1;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = n;
  schablone.name = "L";
  schablone.init = 1;
  lgs->l = nm_matrix.aktion->kreieren(&schablone, NULL);
  lgs->det = 1.0;
  l = lgs->l;                              /* fuer einfacheren Zugriff */
  a = lgs->a->aktion->imVektor(lgs->a,0);
  b = lgs->a->aktion->imVektor(lgs->a,1);
  c = lgs->a->aktion->imVektor(lgs->a,2);
  for (k=1; k<=(n-2); k++) {
    h = a->aktion->amIndex(a, k, 1);
    if (h < EPSILON) 
      return 0;
    lgs->det *= h;
    lkk  = sqrt(h);
    l->aktion->setzen(l, k, k, lkk);          /* l(k,k) = wurzel(a(k)) */ 
    bk   = b->aktion->amIndex(b, k, 1); 
    lk1k = bk/lkk;
    l->aktion->setzen(l, k+1, k, lk1k);       /* l(k+1,k) = b(k)/l(k,k) */
    ck = c->aktion->amIndex(c, 1, 1);
    lnk = ck/lkk;
    l->aktion->setzen(l, n, k, lnk);          /* l(n,k) = c/l(k,k) */
    a->aktion->setzen(a, k+1, 1,              /* a(k+1) = a(k+1)-l(k+1,k)^2 */
      a->aktion->amIndex(a, k+1, 1) 
      - lk1k * lk1k); 
    ck1 = - ck * bk / (lkk * lkk);
    if (k == (n-2))                           /* fallen c(k+1) und b(k+1) */
      ck1 += b->aktion->amIndex(b, k+1, 1);   /* zusammen? */
    c->aktion->setzen(c, 1, 1, ck1);          /* c(k+1) = l(n,k) = siehe */
                                              /* Blatt 03 */
    if (k == (n-2)) 
      b->aktion->setzen(b, k+1, 1, ck1);      /* c(k+1) und b(k+1) identisch */
    a->aktion->setzen(a, n, 1,                /* a(n) = a(n) - l(n,k)^2  */
      a->aktion->amIndex(a, n, 1) - lnk * lnk);
  }
  k = n-1;
  h = a->aktion->amIndex(a, k, 1);
  if (h < EPSILON) 
    return 0;
  lgs->det *= h;
  lkk  = sqrt(h);
  l->aktion->setzen(l, k, k, lkk);          /* l(k,k) = wurzel(a(k)) */ 
  bk   = b->aktion->amIndex(b, k, 1); 
  lk1k = bk/lkk;
  ck   = c->aktion->amIndex(c, 1, 1);
  lnk  = ck/lkk;
  if (fabs(lk1k-lnk)>EPSILON) {
    fprintf(stderr, "Warnung: merkwuerdige Differenz bei der Berechnung.\n");
  }
  l->aktion->setzen(l, n, k, lnk);          /* l(n,k) = c/l(k,k) */
  a->aktion->setzen(a, n, 1,                /* a(n) = a(n) - l(n,k)^2  */
    a->aktion->amIndex(a, n, 1) - lnk * lnk);

  k = n;
  h = a->aktion->amIndex(a, k, 1);
  if (h < EPSILON) 
    return 0;
  lgs->det *= h;
  lkk  = sqrt(h);
  l->aktion->setzen(l, k, k, lkk);          /* l(k,k) = wurzel(a(k)) */ 

  return 1;     /* Erfolg */
}

static int chol06_vorwaertssubstitution(linearesSystem *lgs)
{
  int i, j, n;
  nmmatrixschablone schablone;
  double s;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Vektor c";
  schablone.init = 1;
  lgs->c = nm_matrix.aktion->kreieren(&schablone, NULL);  /* neuer Vektor c */
  for (i=1; i<=n; i++) {
    s = lgs->b->aktion->amIndex(lgs->b, i, 1);           /* s = b(i)        */
    if (i>1) {
      if (i<n) {                                         /* falls i < n     */
        j = i-1;
        s -= lgs->l->aktion->amIndex(lgs->l, i, j)       /* s = s-l(i,j)    */
             * lgs->c->aktion->amIndex(lgs->c, j, 1);    /*       * c(j)    */
      }
      else {                                             /* falls i = n     */
        for (j=1; j<=(i-1); j++) 
          s -= lgs->l->aktion->amIndex(lgs->l, i, j)     /* s = s-l(i,j)    */
                * lgs->c->aktion->amIndex(lgs->c, j, 1); /*      * c(j)     */
      }
    }

    lgs->c->aktion->setzen(lgs->c, i, 1,                 /* c(i) = s/l(i,i) */
      s / lgs->l->aktion->amIndex(lgs->l, i, i));
  }
  return 1;   /* Erfolg */
}

static int chol06_rueckwaertssubstitution(linearesSystem *lgs)
{
  int i, k, n;
  double s;
  nmmatrixschablone schablone;
  n = lgs->l->aktion->breite(lgs->l); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Loesungsvektor x";
  schablone.init = 1;
  lgs->x = nm_matrix.aktion->kreieren(&schablone, NULL);
  for (i=n; i>=1; i--) {                       /* i von Zeile n bis 1  */
    s = lgs->c->aktion->amIndex(lgs->c, i, 1);       /* c(i) +         */
    if (i<n) {
      k = i+1;                                       /* Summe          */
      s += lgs->l->aktion->amIndex(lgs->l, k, i)     /*   l^t(i,k)     */
           * lgs->x->aktion->amIndex(lgs->x, k, 1);  /*   * x(k) von   */
      if (k<n) {
        k = n;
        s += lgs->l->aktion->amIndex(lgs->l, k, i)   /*   l^t(i,k)     */
             * lgs->x->aktion->amIndex(lgs->x, k, 1);/*   * x(k) von   */
      }
    }
    s /= lgs->l->aktion->amIndex(lgs->l, i, i); /* Zeile normalisieren */
    lgs->x->aktion->setzen(lgs->x, i, 1, -s);   /* und x(i) ist freigestellt */
  }
  return 1;   /* Erfolg */
}



/* ----------------------------------------------------------------------
 *
 *  4.) allgemeines CholeskyVerfahren
 */

/*
 * static int chol_lrZerlegungUndDeterminante(linearesSystem *lgs)
 * fehlt noch
 */


static int chol_vorwaertssubstitution(linearesSystem *lgs)
{
  int i, j, n;
  nmmatrixschablone schablone;
  double s;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Vektor c";
  schablone.init = 1;
  lgs->c = nm_matrix.aktion->kreieren(&schablone, NULL);  /* neuer Vektor c */
  for (i=1; i<=n; i++) {
    s = lgs->b->aktion->amIndex(lgs->b, i, 1);        /* s = b(i)           */
    for (j=1; j<=(i-1); j++) 
      s -= lgs->l->aktion->amIndex(lgs->l, i, j)      /* s = s -  l(i,j)    */
           * lgs->c->aktion->amIndex(lgs->c, j, 1);   /*          * c(j)    */
    lgs->c->aktion->setzen(lgs->c, i, 1,              /* c(i) = s/l(i,i)    */
      s / lgs->l->aktion->amIndex(lgs->l, i, i));
  }
  return 1;   /* Erfolg */
}

static int chol_rueckwaertssubstitution(linearesSystem *lgs)
{
  int i, k, n;
  double s;
  nmmatrixschablone schablone;
  n = lgs->l->aktion->breite(lgs->l); 
  schablone.n = n;
  schablone.m = 1;
  schablone.name = "Loesungsvektor x";
  schablone.init = 1;
  lgs->x = nm_matrix.aktion->kreieren(&schablone, NULL);
  for (i=n; i>=1; i--) {                       /* i von Zeile n bis 1  */
    s = lgs->c->aktion->amIndex(lgs->c, i, 1);       /* c(i) +         */
    for (k=(i+1); k<=n; k++) {                       /* Summe          */
      s += lgs->l->aktion->amIndex(lgs->l, k, i)     /*   l^t(i,k)     */
           * lgs->x->aktion->amIndex(lgs->x, k, 1);  /*   * x(k) von   */
    }                                                /* k=i+1 bis n    */
    s /= lgs->l->aktion->amIndex(lgs->l, i, i); /* Zeile normalisieren */
    lgs->x->aktion->setzen(lgs->x, i, 1, -s);   /* und x(i) ist freigestellt */
  }
  return 1;   /* Erfolg */
}


/* ----------------------------------------------------------------------
 * Pivotstrategien
 *   liefern die Nummer der Zeile, die als Pivotzeile benutzt werden soll
 *   oder 0, falls keine Zeile in Frage kommt.
 *   0 kann auch geliefert werden, wenn sich herausstellt, dass die 
 *   gegebene Matrix singulaer ist.
 */

/* Wir stellen dem Benutzer folgende Strategien bereit: */

int diagonalStrategie(matrix *m, int k) 
{
  if (fabs(m->aktion->amIndex(m,k,k)) < EPSILON) 
    return 0;             /* Zeile ablehnen, da Element fast 0 */
  else
    return k;             /* Zeile als Pivotzeile waehlen */
}


int relativeSpaltenMaximumStrategie(matrix *m, int k)
{
  int i, j, n, pk;
  double max, q, s;
  n = m->aktion->breite(m); 
  max = 0.0;
  for (i=k; i<=n; i++) {      /* ab Zeile k nach Pivotzeile suchen */
    s = 0.0;
    for (j=k; j<=n; j++)      /* Zeilensumme berechnen */
      s += fabs(m->aktion->amIndex(m,i,j));
    if (s < EPSILON)          /* Nullzeile? */
      return 0;               /* STOP */
    q = fabs(m->aktion->amIndex(m,i,k)) /s;  /* relativieren */
    if (q > max) {            /* neue, besser geeignete Zeile gefunden */
      max = q;                /* -> merken */
      pk = i;       
    }
  }
  max = fabs(m->aktion->amIndex(m, pk, k));
  if (max < EPSILON)          /* Wenn |m(pk,k)| kleiner epsilon (definiert */
    return 0;                 /* in matrix.h), dann STOP  */ 
  else
    return pk;                /* Pivotzeile zureuckgeben */
} 

